package F5::GlobalSettings;

use strict;
use warnings;
use JSON;
use File::Slurp qw(read_file);
use F5::DbUtils;

our %meta_attributes = map { ($_ => 1 ) } qw(logger
                                             client
                                             file_path_map
                                             allow_cookie_protection_seed_from_file
                                             filename
                                           );

our %CONFIG_ITEM_MAP = (
        maxRequestSize => { config_item_name => "request_buffer_size", factory_value => 10000},
        maxLongRequestSize => { config_item_name => "long_request_buffer_size", factory_value => 10000000},
        maxScanInLongRequest => { config_item_name => "enforce_num_bytes_in_exceeded_request", units => "Bytes"},
        maxConcurrentLongRequests => { config_item_name => "max_concurrent_long_request"},
);

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless({}, $class);
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    $self->{logger} = $params{logger};
    $self->{client} = $params{client};
    $self->{file_path_map} = $params{file_path_map};
    $self->{allow_cookie_protection_seed_from_file} = defined $params{allow_cookie_protection_seed_from_file} ?
        $params{allow_cookie_protection_seed_from_file} : 1;

    if ($params{filename}) {
        my $obj_from_file = $self->load_from_file(filename => $params{filename});
        $self->set_values(%$obj_from_file);
    }

    $self->set_values(%params);
}

#----------------------------------------------------------------------

sub set_values {
    my($self, %params) = @_;

    my $obj_values = {};
    for my $k (keys %params) {
        next if $meta_attributes{$k};
        $obj_values->{$k} = $params{$k};
    }
    $obj_values = F5::ObjUtils::dclone($obj_values);

    F5::ObjUtils::recursively_callback(data_ref => $obj_values,
                                       key_callback => sub { my (%args) = @_;
                                                             return  F5::Utils::Rest::to_camel_case( $args{key} );
                                                         }
                                      );
    F5::ObjUtils::merge_hashes_deeply(source => $obj_values,
                                      target => $self);
}

#----------------------------------------------------------------------

sub apply {
    my($self, %params) = @_;
    my $rv = { completed_successfully => 1 };
    my $prepare_sigs = 0;
    my $do_set_active = 0;
    my $call_flags = F5::ASMConfig::CallFlags->new(%{$params{call_flags} || {}});
    my $skip_set_active = $call_flags->{skip_set_active};
    $call_flags->{skip_set_active} = 1;
    $call_flags->{skip_set_sigs_active} = 1;

    if ($self->{cookieProtection}->{seed} && $self->{cookieProtection}->{digest}) {
        my $msg = "Both seed and digest are configured";
        $self->{logger}->error($msg) if $self->{logger};
        $rv->{completed_successfully} = 0;
        $rv->{failure_msg} = $msg;
        return $rv;
    }
    elsif ($self->{cookieProtection}->{seed}) {
        $self->{logger}->debug("Data protection cookie configuration=$self->{cookieProtection}->{seed}")  if $self->{logger};
        $self->{logger}->debug("Generating new data protection cookie") if $self->{logger};
        $self->{client}->update_data_protection($self->{client}->{session},
                                                $call_flags,
                                                undef,
                                                F5::ASMConfig::DataProtection->new(secure_key => F5::DataProtection::generate_key($self->{cookieProtection}->{seed}))
                                               );
        $rv->{cookie_protection} = $self->{cookieProtection}->{seed};
        $do_set_active = 1;
    }
    elsif ($self->{cookieProtection}->{digest}) {
        $self->{logger}->debug("Data protection cookie configuration=$self->{cookieProtection}->{digest}")  if $self->{logger};
        $self->{logger}->debug("Generating new data protection cookie") if $self->{logger};
        eval {
            $self->{client}->import_data_protection($self->{client}->{session},
                                                    $call_flags,
                                                    undef,
                                                    F5::ASMConfig::DataProtectionImportExport->new(import_text => $self->{cookieProtection}->{digest}),
                                                   );
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            $rv->{failure_msg} = $ex->{error_message};
            return $rv;
        }
        $rv->{cookie_protection} = $self->{cookieProtection}->{digest};
        $do_set_active = 1;
    }

    if ($self->{attackSignatures}->{securityUpdateFile}) {
        my $update_sigfile_rv;
        $self->{logger}->debug("Updating sigfile $self->{attackSignatures}->{securityUpdateFile}") if $self->{logger};
        eval {
            my $get_file_from_link_rv = F5::Utils::File::get_file_from_link(file_path_map => $self->{file_path_map},
                                                                            link => $self->{attackSignatures}->{securityUpdateFile},
                                                                            allow_local_filename => 1);
            my $sigfile = F5::ASMConfig::Sigfile->new(filename => $get_file_from_link_rv->{filepath});
            my $settings = F5::ASMConfig::SigfileUpdateSettings->new(update_type => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                                                                     allow_older_timestamp => 1,
                                                                     im_file => 1);
            $update_sigfile_rv = $self->{client}->update_sigfile($self->{client}->{session}, $call_flags, undef, $sigfile, $settings);
            $self->{sigfile_status_id} = $update_sigfile_rv->{sigfile_status_id};
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            $rv->{failure_msg} = $ex->{error_message};
            return $rv;
        }
        $self->{logger}->debug("Sigfile status id=$self->{sigfile_status_id}") if $self->{logger};
        if (!$update_sigfile_rv->{completed_successfully}) {
            return $update_sigfile_rv;
        }
        if ($self->was_modified(update_file_load_rv => $update_sigfile_rv)) {
            $prepare_sigs = 1;
        }
    }

    if ($self->{botSignatures}->{securityUpdateFile}) {
        my $update_bot_signature_file_rv;
        $self->{logger}->debug("Updating bot signature file $self->{botSignatures}->{securityUpdateFile}") if $self->{logger};
        eval {
            my $get_file_from_link_rv = F5::Utils::File::get_file_from_link(file_path_map => $self->{file_path_map},
                                                                            link => $self->{botSignatures}->{securityUpdateFile},
                                                                            allow_local_filename => 1);
            my $bot_signature_file = F5::ASMConfig::BotSignatureFile->new(filename => $get_file_from_link_rv->{filepath});
            $update_bot_signature_file_rv = $self->{client}->update_bot_signature_file($self->{client}->{session}, $call_flags, undef, $bot_signature_file);
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            if (ref $ex && $ex->{error_message}) {
                $rv->{failure_msg} = $ex->{error_message};
            }
            else {
                $rv->{failure_msg} = $ex;
            }
            return $rv;
        }
        if (!$update_bot_signature_file_rv->{completed_successfully}) {
            return $update_bot_signature_file_rv;
        }
        if ($self->was_modified(update_file_load_rv => $update_bot_signature_file_rv)) {
            $prepare_sigs = 1;
        }
    }

    if ($self->{threatCampaigns}->{securityUpdateFile}) {
        $self->{logger}->debug("Updating threat campaigns file $self->{threatCampaigns}->{securityUpdateFile}") if $self->{logger};
        my $update_tc_rv;
        eval {
            my $get_file_from_link_rv = F5::Utils::File::get_file_from_link(file_path_map => $self->{file_path_map},
                                                                            link => $self->{threatCampaigns}->{securityUpdateFile},
                                                                            allow_local_filename => 1);
            my $threat_campaigns_file = F5::ASMConfig::ThreatCampaignFile->new(filename => $get_file_from_link_rv->{filepath});
            my $settings = F5::ASMConfig::ThreatCampaignFileUpdateSettings->new(allow_older_timestamp => 1);
            $update_tc_rv = $self->{client}->update_threat_campaign_file($self->{client}->{session}, $call_flags, undef, $threat_campaigns_file, $settings);
            $self->{threat_campaign_file_status_id} = $update_tc_rv->{threat_campaign_file_status_id};
        };
        my $ex = $@;
        if ($ex) {
            $rv->{completed_successfully} = 0;
            $rv->{failure_msg} = $ex->{error_message};
            return $rv;
        }
        $self->{logger}->debug("Threat campaign file status id=$self->{threat_campaign_file_status_id}") if $self->{logger};
        if (!$update_tc_rv->{completed_successfully}) {
            return $update_tc_rv;
        }
        if ($self->was_modified(update_file_load_rv => $update_tc_rv)) {
            $prepare_sigs = 1;
        }
    }

    if ($self->{userDefinedSignatures}) {
        $rv->{user_defined_signatures} = $self->load_user_defined_signatures(call_flags => $call_flags);
        $prepare_sigs = 1;
    }

    if ($self->{adaptiveViolationRating} && $self->{adaptiveViolationRating}->{trainingService}) {
        my $adaptive_violation_service = { host => $self->{adaptiveViolationRating}->{trainingService}->{host} || '',
                                           secret => $self->{adaptiveViolationRating}->{trainingService}->{sharedSecret} || '' };
        my $adaptive_violation_service_thrift = F5::ASMConfig::AdaptiveViolationRating->new($adaptive_violation_service);

        my $old_adaptive_violation_service = $self->{client}->get_adaptive_violation_rating($self->{client}->{session}, $call_flags, undef);
        my $does_rec_differ = 0;
        for my $k (keys %$adaptive_violation_service) {
            if ($old_adaptive_violation_service->{$k} ne $adaptive_violation_service->{$k}) {
                $self->{logger}->trace("Adaptive Violation Service differs by attribute $k (old value = $old_adaptive_violation_service->{$k}, new value = $adaptive_violation_service->{$k})") if $self->{logger};
                $does_rec_differ = 1;
            }
        }

        if ($does_rec_differ) {
            $self->{logger}->debug("Updating adaptive violation service") if $self->{logger};
            $self->{client}->update_adaptive_violation_rating($self->{client}->{session}, $call_flags, undef, $adaptive_violation_service_thrift, $adaptive_violation_service_thrift);
            $rv->{adaptive_violation_service} = $adaptive_violation_service;
            $do_set_active = 1;
        }
    }

    if ($self->{enforcerSettings}) {
        $rv->{enforcerSettings} = $self->load_enforcer_settings(call_flags => $call_flags);
        $do_set_active = 1;
    }

    if ($do_set_active || $prepare_sigs) {
        my $dbh = $self->{client}->{dbh};
        if ($prepare_sigs) {
            require F5::NegativeSignatures::SigFile::Update;
            my $negsig_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $self->{client}->{dbh}, logger => $self->{logger});
            $negsig_update->set_sigs_active(asm_config_handler => $self->{client}, skip_set_active => 1);
        }
        if (!$skip_set_active) {
            require F5::SetActive;
            my $set_active_rv = F5::SetActive::set_active(dbh => $dbh,
                                                          expanded_return_value => 1,
                                                          logger => $self->{logger},
                                                          do_update_core => 1,
                                                          send_sync => 0,
                                                          asm_config_handler => $self->{client},
                                                          wait_for_enforcer => $params{wait_for_enforcer},
                                                          job => $self->{job},
                                                         );
            if (!$set_active_rv->{completed_successfully}) {
                return $set_active_rv;
            }
        }
    }

    return $rv;
}

#----------------------------------------------------------------------

sub load_from_file {
    my($self, %params) = @_;

    my $filename = $params{filename};
    my $get_file_rv;
    eval {
        $get_file_rv = F5::Utils::File::get_data_from_link(file_path_map => $self->{file_path_map},
                                                           link => $filename,
                                                           parse_json => 1,
                                                           allow_local_filename => 1);
    };
    my $ex = $@;
    if ($ex) {
        $ex=~ s/ at \S*? line \d+.$//;
        my $error_msg = "Cannot read configuration file '$filename': $ex->{error_message}";
        $self->{logger}->error($error_msg) if $self->{logger};
        die "$error_msg\n";
    }

    my $conf_file_content = $get_file_rv->{data};
    # Normalize config file namespaces and values to camelCase
    F5::ObjUtils::recursively_callback(data_ref => $conf_file_content,
                                       key_callback => sub { my (%args) = @_;
                                                             return  F5::Utils::Rest::to_camel_case( $args{key} );
                                                         }
                                      );

    if (! %$conf_file_content || ! defined $conf_file_content->{wafSettings}) {
        my $error_msg = "No global settings found in '$filename'";
        $self->{logger}->error($error_msg) if $self->{logger};
        die "$error_msg\n";
    }

    if (! $self->{allow_cookie_protection_seed_from_file} && defined $conf_file_content->{wafSettings}->{cookieProtection}->{seed}) {
        $self->{logger}->warn("Unable to read seed from a configuration file") if $self->{logger};
        delete($conf_file_content->{wafSettings}->{cookieProtection}->{seed});
    }

    return $conf_file_content->{wafSettings};
}

#----------------------------------------------------------------------

sub load_user_defined_signatures {
    my($self, %params) = @_;
    my @user_defined_signatures;
    my $sigfile;
    my $seen_tags;
    my $call_flags = $params{call_flags} || F5::ASMConfig::CallFlags->new();
    for my $ref ( @{ $self->{userDefinedSignatures} } ) {
        my $signature_file = F5::Utils::File::get_file_from_link( link => $ref->{'$ref'}, calc_md5 => 1 );
        my $signature_filepath = $signature_file->{filepath};
        my $signature_md5 = $signature_file->{md5};
        $sigfile = F5::Utils::File::load_file( filename => $signature_filepath, json_only => 1 );
        my $error_message;
        if (!$sigfile->{tag}) {
            $error_message = "Missing tag in file '$signature_filepath'";
        }
        elsif ($seen_tags->{$sigfile->{tag}}++) {
            $error_message = "Duplicate tag '$sigfile->{tag}' in file '$signature_filepath'";
        }
        if ($error_message) {
            $self->{logger}->error($error_message) if $self->{logger};
            my $error = {"error_message" => $error_message};
            die $error;
        }
        $self->{logger}->debug("Importing signatures with tag: '$sigfile->{tag}' from file '$signature_filepath'") if $self->{logger};
        my $update_rv      = $self->{client}->update_sigfile(
            $self->{client}->{session},
            $call_flags,
            F5::ASMConfig::Sigfile->new(
                filename            => $signature_filepath,
                flg_is_user_defined => 1,
            ),
            F5::ASMConfig::SigfileUpdateSettings->new(
                allow_older_timestamp => 1,
                im_file               => 0,
                update_type           => F5::ASMConfig::UpdateType::UPDATE_TYPE_ASM_SIGNATURES(),
                validate_requirements => 1,
            )
        );
        my %signature_info = (
            originalFilename => $signature_filepath,
            fileMD5          => $signature_md5,
            tag              => $sigfile->{tag},
        );
        $signature_info{revisionDatetime} = $sigfile->{revisionDatetime} if $sigfile->{revisionDatetime};
        push( @user_defined_signatures, \%signature_info );
    }
    return \@user_defined_signatures;
}

#----------------------------------------------------------------------

sub load_enforcer_settings {
    my($self, %params) = @_;
    my $call_flags = F5::ASMConfig::CallFlags->new(%{$params{call_flags} || {}}, add_on_updating_nonexistent_record => 1);

    my $enforcer_settings_rv = {};
    for my $enforcer_setting (keys %{$self->{enforcerSettings}}) {
        my $config_item = $CONFIG_ITEM_MAP{$enforcer_setting};
        next if (! defined $config_item);
        $config_item->{current_value} = $self->{enforcerSettings}->{$enforcer_setting};
        $config_item->{factory_value} ||= $config_item->{current_value};
        $config_item->{config_item_value_type} = F5::ASMConfig::ConfigItemValueType::CONFIG_ITEM_VALUE_TYPE_NUMBER();
        my $internal_config_item = F5::ASMConfig::InternalConfigItem->new(%{$config_item});
        $enforcer_settings_rv->{$enforcer_setting} = $internal_config_item;
        $self->{client}->update_internal_config_item(
                                                     $self->{client}->{session},
                                                     $call_flags,
                                                     $internal_config_item,
                                                    );

        $self->{logger}->debug("Updated internal config item $config_item->{config_item_name} with value $self->{enforcerSettings}->{$enforcer_setting}");
    }
    my $allow_auto_restart = defined $self->{enforcerSettings}->{allowAutoRestart} ? $self->{enforcerSettings}->{allowAutoRestart} : 0;
    F5::DbUtils::update_row(dbh => $self->{dbh},
                            table_name => "DCC.AUTO_SHUTDOWN",
                            rec => {
                                    flg_auto_shutdown => $allow_auto_restart,
                                   },
                           );
    return $enforcer_settings_rv;
}

#----------------------------------------------------------------------

sub generate_cookie_seed_string {
    my($self, %params) = @_;
    my @chars       = ( 'a' .. 'z', 'A' .. 'Z', '0' .. '9' );
    my $cookie_seed = join '', map { @chars[ rand @chars ] } 1 .. 16;
    return $cookie_seed;
}

#----------------------------------------------------------------------

sub regenerate_data_protection_configuration {
    my($self, %params) = @_;
    my $call_flags = $params{call_flags} || F5::ASMConfig::CallFlags->new();
    my $data_protection = F5::ASMConfig::DataProtection->new(secure_key => F5::DataProtection::generate_key());
    $data_protection->{grace_accepting_interval} = 1;
    $self->{client}->update_data_protection($self->{client}->{session},
                                            $call_flags,
                                            undef,
                                            $data_protection);
}

#----------------------------------------------------------------------

sub was_modified {
    my($self, %params) = @_;
    my $update_file_load_rv = $params{update_file_load_rv};
    for my $k (keys %$update_file_load_rv) {
        if ($k =~ /(added|deleted|updated)_count$/ && $update_file_load_rv->{$k}) {
            return 1;
        }
    }

    return;
}

#----------------------------------------------------------------------

1;
