package F5::ThreatCampaigns;

use strict;
use warnings;

use F5::Logger;
use File::Find;
use F5::DbUtils;
use F5::LockFactory;

use constant THREAT_CAMPAIGN_RULE_SIG_ID_START => 700000000;

#----------------------------------------------------------------------------------------

sub campaign_rules_updated {
    my (%params) = @_;
    require F5::ASMConfig::Entity::PolicyThreatCampaignAttributes;
    require F5::ASMConfig::Entity::PolicyThreatCampaign;
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $match_rule_modified_campaign_ids = $params{match_rule_modified_campaign_ids};


    my @campaign_attributes = F5::ASMConfig::Entity::PolicyThreatCampaignAttributes->retrieve_all(dbh => $dbh,
                                                                logger => $logger,
                                                                get_criteria => {flg_quarantine_threat_campaigns => 1},);
    return unless @campaign_attributes;

    my @policy_ids = map {$_->{policy}->{policy_id}} @campaign_attributes;

    my @search_filters = (
                          F5::ASMConfig::SearchFilter->new(field_name => 'policy_id',
                                                           field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                           field_value_list => \@policy_ids),
                           F5::ASMConfig::SearchFilter->new(field_name => 'threat_campaign__threat_campaign_id',
                                                            field_comparison => F5::ASMConfig::FieldComparison::FIELD_COMPARISON_IN(),
                                                            field_value_list => $match_rule_modified_campaign_ids),
                         );

    F5::ASMConfig::Entity::PolicyThreatCampaign->update_many(dbh => $dbh,
                                                             store_obj => {in_staging => 1},
                                                             call_flags => {search_filters => \@search_filters,},);

}

#----------------------------------------------------------------------------------------

sub add_all_threat_campaigns_to_policy {
    my (%params) = @_;
    my $dbh = $params{dbh};
    my $policy_id = $params{policy_id};
    my $logger = $params{logger};

    my $policy_where_clause = '';
    if ($policy_id ne '*') {
        if (ref $policy_id) {
            $policy_id = join(',',@$policy_id);
        }
        $policy_where_clause = "AND pol.id IN ($policy_id)";
    }

    my $time = time();

    $dbh->do(<<EndQuery, undef, $time);
INSERT INTO PLC.PL_THREAT_CAMPAIGNS
      (policy_id, threat_campaign_id, flg_is_staging, is_inherited, last_updated)
  SELECT pol.id, tc.threat_campaign_id, policy_tca.flg_quarantine_threat_campaigns, policy_tca.is_inherited, ?
    FROM (PLC.PL_POLICIES pol,
          PLC.THREAT_CAMPAIGNS tc)
         LEFT JOIN PLC.PL_THREAT_CAMPAIGNS policy_tc
             ON policy_tc.threat_campaign_id = tc.threat_campaign_id
            AND policy_tc.policy_id = pol.id
         LEFT JOIN PLC.PL_POLICY_THREAT_CAMPAIGN_ATTRIBUTES policy_tca
             ON policy_tca.policy_id = pol.id
   WHERE policy_tc.threat_campaign_id IS NULL
$policy_where_clause
EndQuery

}

#----------------------------------------------------------------------------------------

sub get_campaign_rules_for_policy {
    my (%params) = @_;
    require F5::ASMConfig::ASMConfig;
    my $dbh = $params{dbh};
    my $logger = $params{logger};
    my $policy_id = $params{policy_id};

    my $policy_campaign_rules_sth = $dbh->prepare(<<EndQuery);
SELECT tcmr.*, ptc.flg_is_staging, ptc.flg_enabled, tc.last_updated
  FROM PLC.THREAT_CAMPAIGNS_MATCH_RULES tcmr, PLC.THREAT_CAMPAIGNS tc, PLC.PL_THREAT_CAMPAIGNS ptc
  WHERE ptc.policy_id = ?
  AND tc.threat_campaign_id = ptc.threat_campaign_id
  AND tc.threat_campaign_id = tcmr.threat_campaign_id
  AND tc.status = ?
EndQuery

    my %campaign_rules;
    $policy_campaign_rules_sth->execute($policy_id,
                                        F5::ASMConfig::ThreatCampaignStatus::THREAT_CAMPAIGN_STATUS_ACTIVE(),);
    while (my $match_rule = $policy_campaign_rules_sth->fetchrow_hashref) {
        $campaign_rules{ $match_rule->{threat_campaign_id} } ||= [];
        $match_rule->{sig_id} = get_campaign_rule_sig_id($match_rule->{threat_campaign_match_rule_id});
        push @{ $campaign_rules{ $match_rule->{threat_campaign_id} } }, $match_rule;
    }

    return \%campaign_rules;
}

#----------------------------------------------------------------------------------------

sub get_active_campaign_rules {
    my (%params) = @_;
    require F5::ASMConfig::ASMConfig;
    my $dbh = $params{dbh};
    my $logger = $params{logger};

    my $active_campaign_rules_sth = $dbh->prepare(<<EndQuery);
SELECT DISTINCT tcmr.*, tc.last_updated
  FROM PLC.THREAT_CAMPAIGNS tc, PLC.THREAT_CAMPAIGNS_MATCH_RULES tcmr
  WHERE tc.threat_campaign_id = tcmr.threat_campaign_id
  AND tc.status = ?
EndQuery

    my %campaign_rules;
    my $last_updated = 0;

    $active_campaign_rules_sth->execute(F5::ASMConfig::ThreatCampaignStatus::THREAT_CAMPAIGN_STATUS_ACTIVE(),);

    while (my $match_rule = $active_campaign_rules_sth->fetchrow_hashref) {
        if ($match_rule->{last_updated} > $last_updated) {
            $last_updated = $match_rule->{last_updated};
        }

        $campaign_rules{ $match_rule->{threat_campaign_id} } ||= [];
        $match_rule->{sig_id} = get_campaign_rule_sig_id($match_rule->{threat_campaign_match_rule_id});
        push @{ $campaign_rules{ $match_rule->{threat_campaign_id} } }, $match_rule;
    }

    return {last_updated   => $last_updated,
            campaign_rules => \%campaign_rules};
}

#----------------------------------------------------------------------------------------

sub get_campaign_rule_sig_id {
    my ($threat_campaign_match_rule_id) = @_;
    return $threat_campaign_match_rule_id + THREAT_CAMPAIGN_RULE_SIG_ID_START;
}

#----------------------------------------------------------------------------------------

sub get_update_file_date {
    my(%params) = @_;
    my $flg_any_version = defined $params{flg_any_version} ? $params{flg_any_version} : 0;
    my $dbh = $params{dbh};

    my $where_fields;

    if (!$flg_any_version) {
        $where_fields->{bigip_version} = F5::BigipVersionUtils::bigip_version();
    }
    my $where_clause_rv = F5::DbUtils::construct_where_clause(where_fields => $where_fields);

    my $where_clause = $where_clause_rv->{where_clause};
    my $where_values = $where_clause_rv->{where_values};
    my ($last_update) = $dbh->selectrow_array(<<EndQuery, undef, @{$where_values});
  SELECT threat_campaign_file_timestamp 
    FROM PLC.THREAT_CAMPAIGN_FILE_STATUS 
  WHERE $where_clause
ORDER BY threat_campaign_file_status_id DESC
   LIMIT 1
EndQuery
    return $last_update || 0;
}

#----------------------------------------------------------------------

1;
